#include "C8051F340.h"

#include "testinit.c"
#include <math.h>

unsigned int timedelay=60;

//unsigned int bob;			//random variable for testing
//drive wheel diameter = 1.7 inches
unsigned short int t2of = 0x00;	//number of times timer 2 overflowed
		//incremented in timer 2 interrupt, used for timing ultrasonic
		//input signal.

unsigned int ledperiod=0;
unsigned long int ultrastart=0; //time offset for timer 2 when ultrasonic timing starts
unsigned long int timespan=0;   //duration of ultrasonic pulse from whatever speed timer 2 is

xdata float pi = 3.1416;
xdata float degtorad= 3.1416/180;

xdata unsigned int dist=0;			//last ultrasonic distance reading

xdata short int desired_left_pwm=0;	//desired pwm percentage for left motor, can be + or -
xdata short int desired_right_pwm=0;	//desired pwm percentage for right motor, can be + or -
//desired pwm's to be used by encoder interrupts to maintain proper pwm speeds


xdata short int linear_pwm = 0;				//"linear" pwm sets both left and right pwm to the same value
xdata short int rotational_pwm=0;			//"rotational" pwm is added to left pwm, and subtracted from right pwm
//if rotational_pwm is positive, vehicle should spin somewhat clockwise
//if rotational_pwm is negative, vehicle should spin somewhat counterclockwise

unsigned short int flags=0x00; //initialize flag register
//bit 0 = distance update flag
//flags&=0xFE; resets distance flag
//flags|=0x01; sets distance flag
//flags&0x01 masks off all other flags

//bit 1 = ~50 ms flag
//flags&=0xFD; resets 50 ms flag
//flags|=0x02; sets 50 ms flag
//flags&0x02 masks off all other flags

//bit 2 = 1 ms flag
//flags&=0xFB; resets 1 ms flag
//flags|=0x04; sets 1 ms flag
//flags&0x04 masks off all other flags

//bit 3 = new compass reading flag
//flags&=0xF7; resets compass reading flag
//flags|=0x08; sets compass reading flag
//flags&0x08 masks off all other flags



xdata unsigned int t2count = ~0x0000; //number of increments of timer 2 minus 1
unsigned short int t2overflowcount=0; //number of times timer 2 overflowed, intend to use in
//velocity control routine
xdata unsigned int pcaoffset=0xD11F;	//starting value of PCA register to genetate 1khz PWM
xdata unsigned int pcacount=~0xD11F;
xdata unsigned long int tempvariable=0;


// < port pin used and description of signal>				<c-code for pin>
//ASSUMING PORT 3 PIN 0 IS INFRARED TRANSISTOR SIGNAL		P3_0
__sbit __at (0xB0) IR_beacon;
//assuming port 3 pin 1 is ultrasonic trigger				P3_1

//assuming port 3 pin 2 is left motor DIRECTION				P3_2
//1 = forward, 0 = reverse, check
//assuming port 3 pin 3 is right motor DIRECTION			P3_3
//1 = forward, 0 = reverse, check

__sbit __at (0xB2) left_mtr_forward;
__sbit __at (0xB3) right_mtr_forward;


//assuming port 2 pin 2 is left motor pwm					P2_2
//assuming port 2 pin 3 is right motor pwm					P2_3
//assuming port 2 pin 4 is ultrasonic timing pin			P2_4

//--------------------------------------------------------------------
//                         Steve's variables
//--------------------------------------------------------------------

//unsigned int pwm_left_count = 0;
//unsigned int pwm_right_count = 0;
short int pwm_left_out = 0;
short int pwm_right_out = 0;
int t2difcount_left = 0;
int t2difcount_right = 0;
unsigned int t2overflowcount_left = 0;
unsigned int t2overflowcount_right = 0;


xdata unsigned int truckdirection = 0;
xdata unsigned int bindirection = 0;


//--------------------------------------------------------------------
//
//--------------------------------------------------------------------

//simple control variables
float contrlvar=0.0;
//implememnt simple controller to try to balance the left encoder pulses with right
//in desired ratio
//positive implies more left encoder pulses than right (in a desired ratio)
//negative implies more right encoder pulses than left (in a desired ratio)

//Ex: left wheel to drive at 60% speed, right at 80% speed
//left increment is smallest, set it to 1.0.  Right increment = right speed/left speed = 1.333...
float leftinc=0.0;
float rightinc=0.0;

unsigned long int leftencoderperiod =0;		//time between interrupts, in microseconds
unsigned long int rightencoderperiod = 0;	//time between interrupts, in microseconds
unsigned long int leftencoderoffset=0;		//time from beginning to past interrupt, in microseconds
unsigned long int rightencoderoffset=0;		//time from beginning to past interrupt, in microseconds

//compass variables
unsigned short int bytenumber=0;			//record which byte is currently being recieved
unsigned short int firstbyte=0;				//first recieved byte
unsigned short int secondbyte=0;			//second recieved byte
unsigned int vehicledirection=0;			//recieved orientation from compass


//functions

void compass_enable(void)
{
	P0_5=1;		//prepare compass to operate
	SPIF =0;	//clear all interrupt flags
	WCOL =0;
	MODF =0;
	RXOVRN =0;
	flags&=0xF7; //resets compass reading flag
	
	bytenumber=0;	//reset internal variables
	firstbyte=0;	
	secondbyte=0;
	vehicledirection=0;

	ESPI0=1;	//enable SPI0 interrupt

	P0_5=0;		//activate the compass
}

//*
void compass_disable(void)
{
	P0_5=1;		//deactivate compass
	
	ESPI0=0;	//disable SPI0 interrupt

	SPIF =0;	//clear all interrupt flags
	WCOL =0;
	MODF =0;
	RXOVRN =0;
	
	bytenumber=0;	//reset internal variables
	firstbyte=0;	
	secondbyte=0;
	vehicledirection=0;
}
//*/

void encoder_reset (void){
	int tempcal;
	
	//disable rotary encoder interrupts during update
	EX0=0;
	EX1=0;

	//reset timer 2 overflow count for new controller
	t2overflowcount=0;
	leftencoderoffset=(unsigned long int) (TMR2/4);
	rightencoderoffset=leftencoderoffset;

	//correct to unsigned magnitudes, and update pwm generation
	//direction is controlled in the desired speed routines
	//(do not want controller change drive direction to prevent
	//oscillations)
	if(desired_left_pwm<0)
		pwm_left_out = -desired_left_pwm;
	else
		pwm_left_out = desired_left_pwm;

	if(desired_right_pwm<0)
		pwm_right_out=-desired_right_pwm;
	else
		pwm_right_out=desired_right_pwm;


	tempcal=100-pwm_left_out;
	tempcal=(int)(tempcal*(pcacount/100));

	tempcal = tempcal + pcaoffset + 1;
	if(tempcal<pcaoffset)
		tempcal=0xffff;
	
	PCA0CP0=tempcal;
	
	tempcal=100-pwm_right_out;
	tempcal=(int)(tempcal*(pcacount/100));
	tempcal = tempcal + pcaoffset + 1;
	if(tempcal<pcaoffset)
		tempcal=0xffff;
	PCA0CP1 = tempcal;

	//possibly old controller variable reset
	t2difcount_left = t2count;	
	t2overflowcount_left = t2overflowcount;
	
	//proportional controller variable initialization
	contrlvar=0.0;

	if(pwm_left_out==0||pwm_right_out==0)
	{
		//controller has no effect, just drive at desired pwm
		//cannot maintain ratio of periods for the encoders when one or more motors is stopped
		//for motors
		leftinc=0.0;
		rightinc=0.0;
	}
	else if(pwm_left_out<=pwm_right_out)
	{
		leftinc=1.0;
		rightinc=((float)pwm_right_out)/((float)pwm_left_out);
	}
	else
	{
		rightinc=1.0;
		leftinc=((float)pwm_left_out)/((float)pwm_right_out);
	}
	leftencoderperiod =0;
	rightencoderperiod = 0;

	//re-enable encoders
	EX0=1;
	EX1=1;
}


//updates linear pwm velocity.  Intend to bound between 50% and 80%
//as this range is decently linear.  Will allow rotational pwm to be +/- 10%,
//and controller will allow to adjust it another +/- 20% to correct for desired
//linear and rotational speed
void newlinearpwm(int newpwmpercent)
{
	//if pwm percent is zero, set it to zero
	//if it is higher or lower, bound it between 50 and 80 % pwm
	int rot_corr=1;
	if(newpwmpercent<0)
		rot_corr=-1;

	if(newpwmpercent==0)
	{
		linear_pwm=0;
		desired_left_pwm=rotational_pwm;
		desired_right_pwm=-rotational_pwm;
		if(rotational_pwm>=0)
		{
			left_mtr_forward=1;
			right_mtr_forward=0;
		}
		else
		{
			left_mtr_forward=0;
			right_mtr_forward=1;
		}
	}
	else
	{
		if(rotational_pwm>10)	//correct for if vehicle was rotating in place
			rotational_pwm=10;
		if(rotational_pwm<-10)
			rotational_pwm=-10;
		//resume linear velocity decision loop
		if(newpwmpercent>0)
		{
			//bound between +50 and +80%
			if(newpwmpercent<50)
				newpwmpercent=50;
			if(newpwmpercent>80)
				newpwmpercent=80;
			linear_pwm=newpwmpercent;
			desired_left_pwm = newpwmpercent+rot_corr*rotational_pwm;	//assume this is within bounds of +40 to +90%
			desired_right_pwm = newpwmpercent-rot_corr*rotational_pwm;
			left_mtr_forward=1;		//set motors to drive forward
			right_mtr_forward=1;
		}
		else
		{
			if(newpwmpercent>-50)
				newpwmpercent=-50;
			if(newpwmpercent<-80)
				newpwmpercent=-80;

			linear_pwm=newpwmpercent;
			desired_left_pwm = newpwmpercent+rot_corr*rotational_pwm;	//assume this is within bounds of +40 to +90%
			desired_right_pwm = newpwmpercent-rot_corr*rotational_pwm;
			left_mtr_forward=0;		//set motors to drive backward
			right_mtr_forward=0;
		}
	}
	encoder_reset();
}


void newrotationalpwm(int newpwmpercent)
{
	//correction flag for if motors are going backward
	//when the motors are going backward, the vehicle will rotate
	//counter-clockwise if pwm is just directly added to it
	int rot_corr=1;
	if(linear_pwm<0)
		rot_corr=-1;

	if(newpwmpercent==0)
	{
		rotational_pwm=0;
		desired_left_pwm=linear_pwm;
		desired_right_pwm=linear_pwm;
	}
	else if(linear_pwm==0)//if vehicle is not driving forward
	{
		if(newpwmpercent>0)
		{//rotate clockwise between +40 and +90 % pwm
			if(newpwmpercent>90)
				newpwmpercent=90;
			if(newpwmpercent<40)
				newpwmpercent=40;
			rotational_pwm=newpwmpercent;
			desired_left_pwm=newpwmpercent;
			desired_right_pwm=-newpwmpercent;
			left_mtr_forward=1;
			right_mtr_forward=0;
		}
		else
		{//rotate counter-clockwise between -40 and -90% pwm
			if(newpwmpercent<-90)
				newpwmpercent=-90;
			if(newpwmpercent>-40)
				newpwmpercent=-40;
			rotational_pwm=newpwmpercent;
			desired_left_pwm=newpwmpercent;
			desired_right_pwm=-newpwmpercent;
			left_mtr_forward=0;
			right_mtr_forward=1;
		}
	} 
	else 
	{
		//bound between +/- 10% pwm
		if(newpwmpercent>10)
			newpwmpercent=10;
		if(newpwmpercent<-10)
			newpwmpercent=-10;
		rotational_pwm=newpwmpercent;
		
		desired_left_pwm=linear_pwm+rot_corr*newpwmpercent;
		desired_right_pwm=linear_pwm-rot_corr*newpwmpercent;
		//motor directions should already have been set in linear
		//pwm routine, will not change them here.
	}
	encoder_reset();
}

//REMEMBER: This is occurring inside an interrupt, may need to modify if it gets too long.
void distupdate(int newdist)
{
	//removed distance offset for bucket distance and whatnot, since some of the times it is there,
	//and some of the times it is not (arm up => no threshold, arm down => threshold)
	//assume programmer will compensate on his or her own for this at this time.
	//note: distance = centimeters

	//set flag for new distance update
	flags|=0x01;
	//filtering
	
	//no filtering at this time... sensor does not appear to have much noise

	//update output
	dist=newdist;
}


//wait time function, not tested for times longer than 12-13 seconds
//uses 50 ms flag
void waitsec(unsigned int seconds) 
{
	unsigned int exitcount=seconds*20;
	unsigned int timeout=0;
	while(1)
	{
		if(flags&0x02>0)
		{
			flags&=0xFD;		//reset 50 ms flag
			timeout++;
			if(timeout>=exitcount) //20 pulses per second * 10 seconds = 200 pulses
			{
				break;
			}
		}
	}
}



void waitmsec(unsigned long int milliseconds) 
{

	while(milliseconds>0)
	{
		if(flags&0x04>0)
		{
			flags&=0xFB; //resets 1 ms flag
			milliseconds--;
		}
	}
}


//try to use a fast stopping routine to reduce rotational and linear overshoot
//reverse direction on drive motors, wait for a few subsequent pwm signals to help
//remove vehicle inertia, then stop and reset all variables
//
//also, reset arm and bucket motor signal to 0000 to stop both.
void vehiclestop(void)
{
//plan: reverse directions on drive motors for a few pwm updates, then stop them and return
	unsigned int delaytime=10;

	left_mtr_forward=!left_mtr_forward;		//reverse direction
	right_mtr_forward=!right_mtr_forward;

	P4=P4&0xF0;			//stop arm and bucket motors
	flags&=0xFB;		//reset 1 ms flag
	//note: should be replaced with call to waitmsec(delaytime);
	while(1)
	{	//wait for a few cycles of PWM signal before stopping
		//leave magnitude the same, since it makes the most sense
		//(if going 100% forward, go 100% backward for a period of time
		//versus going backward 0% for the same period of time
		if(flags&0x04)		//check 1 ms flag
		{
			flags&=0xFB;	//reset 1 ms flag
			delaytime--;
			if(delaytime==0)
			{
				break;
			}
		}
	}
	
	PCA0CP0=0xFFFF;	//stop motors
	PCA0CP1=0xFFFF;	

	desired_left_pwm=0;	//set all internal velocity states to zero
	desired_right_pwm=0;
	linear_pwm=0;
	rotational_pwm=0;
	encoder_reset();
}



//desire function to control arm and bucket motors
//process:
//stop drive motors => drive motors produce enough noise to cause false triggers of
//the sensor.  There may also be issues of noise if arm and bucket motors running 
//simultaneously as well.  Hardware change: Add 4 pullup resistors to vehicle sensors
//current software correction: wait for signal to be low for ~1 second
//possible software correction: check that the majority of the past second has been high,
//or that the pulses caused by noise are no longer than a certain time
//inputs: 2 integers, bucket_raise and arm_raise
//for bucket_raise: if it is 0, nothing happens, if it is < 0, the bucket is lowered
//and if it is > 0, the bucket is raised.
//for arm_raise: if it is 0, nothing happens, if it is < 0, the arm is lowered
//and if it is > 0, the arm is raised
//
//NOTE: P4 is only byte addressable, and motor control signals use port 4.
//need to write the whole byte to the port in order for this to work properly.
//will take port values, and use logic operations to get desired signals
//(to preserve port 4 high byte information)
//ADDITIONAL NOTE: Bucket grinds against drive tracks if the bucket AND arm are
//down at the same time.  Code prevents this operation in one function call
void bucketraise_armraise(short int bucket_raise, short int arm_raise)
{
	unsigned int specialsensorcount=0;
	unsigned short int tempp4=0;
	//could save linear and rotational pwm and restore after process complete...
	
	
	if((bucket_raise<0)&&(arm_raise<0))
	{	
		arm_raise=0;	//disable arm motor reaction
	}
	vehiclestop();

	if(bucket_raise>0)
	{//raise bucket until sensor is hit
		tempp4=P4;
		tempp4|=0x04;	//set bucket raise
		tempp4&=0xf4;	//clear bucket lower and arm signals
		P4=tempp4;		//reset all at once
		while(1)
		{
			while(P3_6==1)
			{;}
			specialsensorcount=1;
			while(P3_6==0)
			{
				if(flags&0x02>0)
				{
					flags&=0xFD;
					specialsensorcount++;
					if(specialsensorcount>=20)
					{
						break;
					}
				}

			}
			if(specialsensorcount>=20)
				break;
		}

	}
	
	if(bucket_raise<0)
	{//lower bucket until sensor is hit
		tempp4=P4;
		tempp4|=0x08;	//set bucket lower
		tempp4&=0xf8;	//clear bucket raise and arm signals
		P4=tempp4;		//reset all at once
		while(1)
		{
			while(P3_7==1)	//wait for sensor to go low
			{;}
			specialsensorcount=1;	//reset timing variable
			while(P3_7==0)	//time how long sensor is low
			{
				if(flags&0x02>0)	//use 50 ms flag for longer timing
				{
					flags&=0xFD;
					specialsensorcount++;
					if(specialsensorcount>=20)	//check for longer than 1 second of 0's
					{
						break;
					}
				}

			}
			if(specialsensorcount>=20)
				break;
		}

	}

	if(arm_raise>0)
	{//raise arm until sensor is hit
		tempp4=P4;
		tempp4|=0x01;	//set arm raise
		tempp4&=0xf1;	//clear arm lower and bucket signals
		P4=tempp4;		//reset all at once
		while(1)
		{
			while(P3_4==1)
			{;}
			specialsensorcount=1;
			while(P3_4==0)		//noisy signal, wait for approx 2 milliseconds of zero readings to stop
			{
				if(flags&0x02>0)
				{
					flags&=0xFD;

					specialsensorcount++;
					if(specialsensorcount>=20)
					{
						break;
					}
				}

			}
			if(specialsensorcount>=20)
				break;
		}
	}
	
	if(arm_raise<0)
	{//lower arm until sensor is hit
		tempp4=P4;
		tempp4|=0x02;	//set arm lower
		tempp4&=0xf2;	//clear arm raise and bucket signals
		P4=tempp4;		//reset all at once
		while(1)
		{
			while(P3_5==1)
			{;}
			specialsensorcount=1;
			while(P3_5==0)		
			{
				if(flags&0x02>0)
				{
					flags&=0xFD;
					specialsensorcount++;
					if(specialsensorcount>=20)
					{
						break;
					}
				}

			}
			if(specialsensorcount>=20)
				break;
		}
	}


}




short int drivetobeacon(unsigned short int desired_dist, unsigned short int beacon_off_time, unsigned short int ir_track)
{
	unsigned int curdist=0;
	unsigned int distdiff=0;
	unsigned short int pwmpercent=0;
	unsigned int irbeacontimeout=0;
	flags&=0xFE;		//reset distance flag

	while(1)//drive to wall
	{
		while(flags&0x01==0)		//wait for new distance reading	
		{;}
		flags&=0xFE;				//reset flag
		curdist=dist;				//read dist
		
		if(curdist>=desired_dist) //vehicle is far from object
		{
			//start of vehicle too far away

			distdiff=curdist-desired_dist;
			if(distdiff>3)		//allow tolerance of 3 cm from target
			{
				//try to have velocity proportional to distance
				//with pwm between 50 and 80% pwm with distdiff between 4 and 50
				//limit to 50% to hopefully minimize non-linearities...
				
				if(distdiff>50)
					distdiff=50;
				pwmpercent=distdiff*2;
				if(pwmpercent<50)
					pwmpercent=50;
				if(pwmpercent>80)
					pwmpercent=80;

				newlinearpwm(pwmpercent);
			}
			else
			{
				break;
			}
			//end of vehicle too far
		}

		else
		{
			//start of vehicle too close

			distdiff=desired_dist-curdist;
			if(distdiff>3)		//allow tolerance of 3 cm from target
			{
				//try to have velocity proportional to distance
				//with pwm between 50 and 100% pwm with distdiff between 2 and 25
				//limit to 50% to hopefully minimize non-linearities...

				if(distdiff>50)
					distdiff=50;
				pwmpercent=distdiff*2;
				if(pwmpercent<50)
					pwmpercent=50;
				if(pwmpercent>80)
					pwmpercent=80;
				newlinearpwm(-1*(int)pwmpercent); //tell vehicle to drive backwards
			}
			else
			{
				break;
			}
			//end of vehicle too close

		}

		//verify IR transistor is still visible
		if(ir_track==1)
		{
			if(IR_beacon==0)
			{
				//reset parameters
				irbeacontimeout=0;
				newrotationalpwm(0);
			}
			else
			{
				//turn slightly to look for IR beacon, counterclockwise first, then clockwise
				//process idea: turn counterclockwise (opposite of the direction to seek IR led) first
				//for some time (50 ms maybe?), then turn clockwise for twice the time of turning counterclockwise
				//(half to get back to zero, half to look to the right of the vehicle).  If IR beacon is not found, 
				//stop and call the IR location routine again...
				irbeacontimeout++;
				if(irbeacontimeout==beacon_off_time)
				{
					flags&=0xFB;			//reset 1 ms flag
					newrotationalpwm(-10);
				}
				else if(flags&0x04>0)	//check for 1 ms flag
				{
					flags&=0xFB;		//reset flag
					irbeacontimeout++;
					if(irbeacontimeout>(70+beacon_off_time))	//end time for turning counterclockwise
						newrotationalpwm(10);
					if(irbeacontimeout>(270+beacon_off_time))	//end time for turning clockwise 
					{	//note: time is at least 2*counterclockwise time + clockwise time
						//allowing more than 2*counterclockwise time due to momentum
						//give up "small" looking, and start process again
						vehiclestop();
						return(-1);
					}
				}
			}
		}

	}

	return(0);

}




void degreeheading(int degreeface, short int tolerance)
{
	xdata int absoluteheading=0;
	xdata int deglimit=0;
	xdata int deltaangle=0;
	xdata short int directioncorrection=1;

	vehiclestop();
	waitmsec(200);	//let magnetic interference die down
	flags&=0xF7; //resets compass reading flag
	//look for two compass readings within +/- 1 degree (unchecked for wrap around)
	while(1)
	{
		while(flags&0x08==0)
		{;}	//wait for new compass reading
		absoluteheading=vehicledirection;
		flags&=0xF7; //resets compass reading flag
		while(flags&0x08==0)
		{;}	//wait for new compass reading
		if(absoluteheading>=(int)(vehicledirection)-1 & absoluteheading<=vehicledirection+1)
			break;
	}
	flags&=0xF7; //resets compass reading flag
	//current compass value is in absoluteheading

	//check cases:
	//degreeface = 271, heading = 90, rotate counterclockwise
	//     deltaangle = 181, subtract 360 degrees, delta angle=-179 degrees.  rotate counterclockwise
	//degreeface=269, heading = 90, rotate clockwise
	//     deltaangle = 179, leave as is, rotate clockwise
	//degreeface=90, heading=271, rotate clockwise
	//     deltaangle = -181, add 360, delta angle = 179, rotate clockwise
	//degreeface=90, heading = 269; rotate counterclockwise
	//     deltaangle = -179, leave as is, rotate counterclockwise
	//degreeface=359, heading = 1, rotate counterclockwise
	//     deltaangle = 358, subtract 360, deltaangle=-2, rotate counterclockwise
	//degreeface = 1, heading = 359, rotate clockwise
	//     deltaangle = -357, add 360, deltaangle=3, rotate clockwise
	//test cases rotate correct direction

	deltaangle=degreeface-absoluteheading;
	if(deltaangle>180)
		deltaangle-=360;
	if(deltaangle<-180)
		deltaangle+=360;
	if(deltaangle>tolerance)
	{
		directioncorrection=1;
	}
	else if(-deltaangle>tolerance)
	{
		directioncorrection=-1;
		deltaangle=-deltaangle;
	}
	else {directioncorrection=0;}


	while(1)
	{
		newrotationalpwm(30*directioncorrection);
		if(deltaangle<=15)
		{
			waitmsec(50);
			vehiclestop();
			waitmsec(200);
			flags&=0xF7; //resets compass reading flag
			while(flags&0x08==0)
			{;}	//wait for new compass reading
			flags&=0xF7; //resets compass reading flag
			while(flags&0x08==0)
			{;}	//wait for new compass reading
			flags&=0xF7; //resets compass reading flag
			absoluteheading=vehicledirection;

			deltaangle=degreeface-absoluteheading;
			if(deltaangle>180)
				deltaangle-=360;
			if(deltaangle<-180)
				deltaangle+=360;
			if(deltaangle>tolerance)
			{
				directioncorrection=1;
			}
			else if(deltaangle<-tolerance)
			{
				directioncorrection=-1;
				deltaangle=-deltaangle;
			}
			else 
			{
				break;
			}
		}
		else
		{
			while(1)
			{
				if(rotational_pwm==30*directioncorrection)
				{}
				else
				{
					newrotationalpwm(30*directioncorrection);
				}

				absoluteheading=vehicledirection;

				deltaangle=degreeface-absoluteheading;
				if(deltaangle>180)
					deltaangle-=360;
				if(deltaangle<-180)
					deltaangle+=360;
				if(deltaangle>15)
				{
					directioncorrection=1;
				}
				else if(-deltaangle>15)
				{
					directioncorrection=-1;
					deltaangle=-deltaangle;
				}
				else //leave
				{
					if(deltaangle<0)
						deltaangle=-deltaangle;
					vehiclestop();
					break;
				}
			}
		}
	}
}



//function to look for infrared beacon.  Stop the vehicle, rotate
//in a circle, and try to locate the desired IR source
//targetloc ==0, find next clockwise infrared beacon
//targetloc > 0, find load material
//		load material beacon flashes at XX Hz
//targetloc < 0, find truck
//		truck beacon flashes at YY Hz
//note: infrared circuitry does not appear to work much faster than ~10 Hz
//may try to come up with better detection circuitry
//
//returns 0 if desired target found
// ""	  -1 if target could not be found
//
//assuming load blinks at 10 Hz, and truck blinks at 5 Hz initially
//and they have high enough duty cycles to locate
//allow period to vary +/- 40 ms from desired period, to handle inconsistencies in timing, sensors,
//IR led timing signals, etc.
short int locateIRbeacon(short int targetloc)
{
	xdata unsigned int compasstolerance=20;
	xdata int highthresh=0;
	xdata int lowthresh=0;

	unsigned int timeout=0;		//time out variable used for looking for load, and for non-changing IR light
	//set timeout for the non-changing IR light before going through the target cases if needed to change
	unsigned short int remainingtargets=3;	//number of infrared targets to check
	unsigned short int waitcount=120;	//look for ~6 seconds (number of 50 ms flags to clear)
	unsigned int loadperiod=100;	//period of load IR beacon, in ms
	unsigned int truckperiod=200;	//period of truck IR beacon, in ms
	unsigned int countvar=0;	//counter variable for target identification


	if(targetloc>0)
	{
		highthresh=bindirection+compasstolerance;
		lowthresh=bindirection-compasstolerance;
		if(highthresh>360)
			highthresh=highthresh-360;
		if(lowthresh<0)
			lowthresh+=360;
	}
	else if(targetloc<0)
	{
		highthresh=truckdirection+compasstolerance;
		lowthresh=truckdirection-compasstolerance;
		if(highthresh>360)
			highthresh=highthresh-360;
		if(lowthresh<0)
			lowthresh+=360;
	}



	while(1)
	{//desired target location loop
		vehiclestop();			//stop vehicle
		//check if the number of IR signals found has exceeded allowed limit
		if(remainingtargets==0)
		{	//too many IR targets found while desired target not found, return error
			return(-1);
		}
		newrotationalpwm(30);	//rotate clockwise slowly
		timeout=0;				//reset timeout variable
		countvar=0;				//reset count variable
		ledperiod=0;


		while(1)
		{//IR beacon locate loop
			if(IR_beacon==0)
			{
				break;
			}
			else
			{
				if(flags&0x02>0)
				{
					flags&=0xFD;		//reset 50 ms flag
					timeout++;
					if(timeout>=waitcount)
					{
						break;
					}
				}
			}
		}//end of IR beacon locate loop
		
		vehiclestop(); //stop vehicle
		
		//check for timing out of previous loop
		if((timeout>=waitcount))
		{
			//could not locate any target, return error
			//could replace in future with semi-intelligent drive routine
			return(-1);
		}



		remainingtargets--;		//past section where code can time out
								//decrease number of false IR targets allowed

		if(dist>45)
			drivetobeacon(40,30,0);

		flags&=0xF7; //resets compass reading flag
		waitmsec(250);
		if(vehicledirection>=lowthresh|vehicledirection<=highthresh)
		{	//vehicle may be within bounds
			if(highthresh>lowthresh)
			{
				if(vehicledirection>=lowthresh&vehicledirection<=highthresh)
				{
					//vehicle is in the right location, do nothing
				}
				else	//vehicle out of bounds, restart loop
				{
					newrotationalpwm(30);
					waitmsec(250);
					vehiclestop();
					continue;
				}
			}
		}
		else
		{
			newrotationalpwm(30);
			waitmsec(250);
			vehiclestop();
			continue;
		}

		countvar=1000;	//highest period IR signal + tolerance
		timeout=1000;	//highest period IR signal + tolerance

		//IR_beacon should be 0 at this point, 
		//check to see which case function is looking for
		if(targetloc==0)
		{	//looking for any IR target, and one was found, so exit loop
			return(0);
		}
		else 
		{	//wait for the start of a new cycle, and eliminate DC IR light
			while(IR_beacon==0)
			{
				if(flags&0x04>0)		//check 1 ms flag
				{
					flags&=0xFB;	//reset 1 ms flag
					timeout--;
					if(timeout==0)
						break;
				}
			}
			//IR_beacon should be 1 now, or timed out from dc source
			if(timeout==0)
			{	//no change in sensor in desired time, must be (virtually) constant IR source
				//like the sun, or lights, or etc... rotate a little bit then redo loop
				newrotationalpwm(30);
				waitmsec(250);
				vehiclestop();
				continue;				//restart loop
			}
			
			//measure period of one transition (consists of high and low cycle)
			ledperiod=0;
			while(IR_beacon==1)
			{
				if(flags&0x04)		//check 1 ms flag
				{
					flags&=0xFB;	//reset 1 ms flag
					ledperiod++;		//increase ms count time
				}
				if(ledperiod>countvar)	//assume no frequency below .5 Hz...
					break;
			}
			while(IR_beacon==0)
			{
				if(flags&0x04)		//check 1 ms flag
				{
					flags&=0xFB;	//reset 1 ms flag
					ledperiod++;		//increase ms count time
				}
				if(ledperiod>countvar)	//assume no frequency below .5 Hz...
					break;
			}

			if(targetloc>0)
			{//looking for load material
				if((ledperiod>(loadperiod+40))|(ledperiod<(loadperiod-40)))
				{//target is not desired target, repeat loop again
					newrotationalpwm(30);
					waitmsec(250);
					vehiclestop();
					continue;
				}
				break;
			}
			else
			{
				//looking for truck
				if((ledperiod>(truckperiod+40))|(ledperiod<(truckperiod-40)))
				{//target is not desired target, repeat loop again
					newrotationalpwm(30);
					waitmsec(250);
					vehiclestop();
					continue;
				}
				break;
			}
		}
	}
 	return(0);
}




void main (void) {	
	unsigned short int distancecount=0;
	unsigned short int prevdistance=0;
	unsigned short int pwmpercent=0;
	unsigned long int pcavalues=0;

	short int returnval=0;
	short int target=1;		//start looking for load
	unsigned int loadpwmtest=0;


	P4=0x01;	//set vehicle to raise arm
	PCA0CP1=0xFFFF; //stop vehicle pwm
	PCA0CP0=0xFFFF;

	Init_Device();
	TMR2RL   = ~ t2count;	//overwrite timer 2 reload so config cant screw it up
	PCA0MD &= ~0x40;		//overwrite pca mode
	IT0=1;	//configure external interrupts to edge triggered
	IT1=1;	//configure external interrupts to edge triggered
	EX0=1;
	EX1=1;
	vehiclestop();


	bucketraise_armraise(1,-1); //raise bucket then lower arm
	P4&=0xF1;		//set arm raise manually for short duration
	P4|=0x01;
	waitsec(2);		//wait a short time, so arm is off ground
	P4&=0xF0;		//stop arm motor


	compass_enable();



	while(P2_0==1)
	{;}
	waitsec(1);
	while(1)
	{
		while(flags&0x08==0)
		{;}	//wait for new compass reading
		bindirection=vehicledirection;
		flags&=0xF7; //resets compass reading flag
		while(flags&0x08==0)
		{;}	//wait for new compass reading
		if(bindirection>=(int)(vehicledirection)-1 & bindirection<=vehicledirection+1)
			break;
	}

	flags&=0xF7; //resets compass reading flag
	waitsec(2);
	while(P2_0==1)
	{;}
	waitsec(1);
	while(1)
	{
		while(flags&0x08==0)
		{;}	//wait for new compass reading
		truckdirection=vehicledirection;
		flags&=0xF7; //resets compass reading flag
		while(flags&0x08==0)
		{;}	//wait for new compass reading
		if(truckdirection>=(int)(vehicledirection)-1 & truckdirection<=vehicledirection+1)
			break;
	}
	truckdirection = vehicledirection;
	flags&=0xF7; //resets compass reading flag
	waitsec(1);


	//main loop
	while(1)
	{
//*
		//begin IR location routine
		if(target==1)
			degreeheading(bindirection,4);

		if(target==-1)
			degreeheading(truckdirection,4);

		returnval=locateIRbeacon(target);

		if(returnval==-1)
		{
			waitsec(4);	//wait and repeat main loop
			continue;
		}

		if(target==1)
			degreeheading(bindirection,4);

		if(target==-1)
			degreeheading(truckdirection,4);


		//end of IR location
//*/
		//begin drive to object routine
		//drive to object
		
		returnval=drivetobeacon(40,50,1);
		if(returnval==-1)
		{
			waitsec(4);
			continue;
		}

		if(target==1)
		{//load material routine
			
			bucketraise_armraise(1,-1); //raise bucket then lower arm

//-----------------------------------------------
//          Test load function
//-----------------------------------------------
			distancecount=0;
			prevdistance=0;
			newlinearpwm(50);
			waitmsec(400);
			prevdistance=dist;
//bit 0 = distance update flag
//flags&=0xFE; resets distance flag
//flags|=0x01; sets distance flag
//flags&0x01 masks off all other flags
			flags&=0xFE; //resets distance flag
			loadpwmtest = rightencoderperiod;
			loadpwmtest += leftencoderperiod;

			while (1){
				if (1.05 * loadpwmtest < rightencoderperiod + leftencoderperiod) 
					break;
				if(flags&0x01==1) //new distance reading
				{
					flags&=0xFE;
					if(dist<prevdistance)
					{
						prevdistance=dist;
						distancecount=0;
					}
					else if (dist==prevdistance)
					{
						distancecount++;
						if(distancecount>10)
							break;
					}
				}
			}
			P4&=0xF1;
			P4|=0x01;
			newlinearpwm(-50);
			waitmsec(200);
			newlinearpwm(0);
			waitsec(1);
			waitmsec(500);
			newlinearpwm(-50);
			waitsec(1);
//			waitmsec(500);
			P4&=0xF0;
			vehiclestop();
			newrotationalpwm(40);
			waitsec(1);
			vehiclestop();
			target = -1;

//----------------------------------------------
		}
		else if (target==-1)
		{
			bucketraise_armraise(1,1);		//raise bucket and arm all the way up for dumping
			//drivetoobject(5);				//drive close to object, since bucket is not in way
			returnval=drivetobeacon(6,20,0);
			bucketraise_armraise(-1,0);		//dump material
			//drivetoobject(30);			//back up from object to look for bin
			returnval=drivetobeacon(40,20,0);
			bucketraise_armraise(1,-1);		//tip up bucket and lower arm
			
			P4&=0xF1;		//set arm raise manually for short duration
			P4|=0x01;
			waitsec(3);		//wait a short time, so arm is off ground
			P4&=0xF0;		//stop arm motor

			newrotationalpwm(30);
			waitsec(2);
			vehiclestop();

			target=1;		//target material
		}


	}//end of main while loop

}//end of main function

void Timer2_ISR (void) interrupt 5
{
   unsigned short int i = 0;
   t2overflowcount++;
   t2of++;
   TF2H = 0;
   if (t2of > 0x02 ) {
   t2of=0x00;
   flags|=0x02;	//set ~50 ms flag
   P3_1 = 1;
   for (i = 0; i < 45; i++); //manually wait time for ultrasonic trigger generation
   P3_1 = 0;}
 
}


void PCA_ISR (void) interrupt 11
{
	unsigned short int resetvar;
	if(CCF0==1)
	{
		CCF0=0;
		//module 0 interrupt
	}
	else if(CCF1==1)
	{
		CCF1=0;
		//module 1 interrupt
	}
	else if(CCF2==1)
	{
		CCF2=0;
		//module 2 interrupt
		if( P2_4==1)
		{
			ultrastart = (long int)TMR2;
			ultrastart = ultrastart - ~t2count;
			ultrastart=(long int) ultrastart/4;
			tempvariable=(long int)t2count/4;
			tempvariable=(long int)tempvariable*t2of;
			ultrastart = ultrastart + tempvariable; //correct for delays
		}
		else
		{
			timespan = TMR2;
			timespan = timespan - ~t2count; //subtract off offset from reload time
			timespan = (long int)timespan/4;
			tempvariable=(long int)t2count/4;
			tempvariable=(long int)tempvariable*t2of;

			timespan = timespan + tempvariable;  //add on how many overflow counts occurred
			timespan=timespan-ultrastart;	   	//correct for delay
//move into function, for potential filtering
			flags|=0x01;
			dist=(int)(timespan/58);
		}

	}
	else if(CCF3==1)
	{
		CCF3=0;
		//module 3 interrupt
//		resetvar=SWRSF;    // software reset
//		resetvar|=0x01;
		RSTSRC|=0x10;
	}
	else if(CCF4==1)
	{
		CCF4=0;
		//module 4 interrupt
	}
	else 
	{
		PCA0 = pcaoffset; // + PCA0; //removed adding on offset of PCA to hopefully make pwm more reliable
		//may result in some "jitter" in the 1 khz frequency, but only by probably about 40-50 instructions max, or ~1 us
		//will assume negligible impact
		flags|=0x04;	//set 1 ms flag
		CF = 0;
	}
}       

//simple velocity controller

void encoder_pwmupdate(void)
{
	unsigned int templeftpwm, temprightpwm;
	short int pwmlimit=20;		//how far +/- from desired pwm vehicle can go
	unsigned short int gainfactor=5;
	short int deltachange=0;

	EX0=0;
	EX1=0;
	
	if(desired_left_pwm<0)
		desired_left_pwm=-desired_left_pwm;
	if(desired_right_pwm<0)
		desired_right_pwm=-desired_right_pwm;
	
	if(desired_left_pwm<40)
		desired_left_pwm=40;
	if(desired_right_pwm<40)
		desired_right_pwm=40;

	if(contrlvar>50.0)
		contrlvar=50.0;
	if(contrlvar<-50.0)
		contrlvar=-50.0;

	if(contrlvar<1.5 && contrlvar>-1.5)
		deltachange=0;
	else
		deltachange=(short int)(gainfactor*contrlvar);

	if(deltachange>2*pwmlimit)
		deltachange=2*pwmlimit;
	if(deltachange<-2*pwmlimit)
		deltachange=-2*pwmlimit;

	//update corresponding motor pwm percentage (bound between 30 and 100%
	if(deltachange>=0)
	{  
		//left motor too fast and/or right motor too slow
		//try to increase right motor pwm, if unable use all change, decrease left motor pwm
		pwm_right_out= desired_right_pwm + deltachange;
		if(pwm_right_out>desired_right_pwm+pwmlimit)
		{
			deltachange=pwm_right_out-desired_right_pwm - pwmlimit; //remaining change
			pwm_right_out=desired_right_pwm + pwmlimit;
			pwm_left_out=desired_left_pwm-deltachange;
			if(pwm_left_out<desired_left_pwm-pwmlimit)
			{
				pwm_left_out=desired_left_pwm-pwmlimit;
			}
		}
	}
	else
	{
		pwm_left_out=desired_left_pwm - deltachange;//(remember, deltachange is negative)
		if(pwm_left_out>desired_left_pwm+pwmlimit)
		{
			deltachange=pwm_left_out-desired_left_pwm - pwmlimit; //remaining change
			pwm_left_out=desired_left_pwm+pwmlimit;
			pwm_right_out=desired_right_pwm-deltachange;
			if(pwm_right_out<desired_right_pwm-pwmlimit)
			{
				pwm_right_out=desired_right_pwm-pwmlimit;
			}
		}
	}

	if(pwm_left_out<40)
		pwm_left_out=40;
	if(pwm_left_out>100)
		pwm_left_out=100;
	if(pwm_right_out<40)
		pwm_right_out=40;
	if(pwm_right_out>100)
		pwm_right_out=100;
	
	//to calculate pwm signal, invert percentage, multiply it by the
	//total number of increments per pwm cycle, divide by 100 to cancel percentage
	//add to offset, then add 1 for the case of 100% pwm
	//(without the +0, the PCA counter never increments to the PCA value, and the
	//signal sometimes transitions, and sometimes doesnt)
	templeftpwm=(100-pwm_left_out)*(int)(pcacount/100)+pcaoffset+1;
	temprightpwm=(100-pwm_right_out)*(int)(pcacount/100)+pcaoffset+1;
	if(templeftpwm<pcaoffset)//check for overflow from 0% case
		templeftpwm=0xffff;
	if(temprightpwm<pcaoffset)
		temprightpwm=0xffff;
	//update pca's near simultaneously
	PCA0CP0=templeftpwm;
	PCA0CP1=temprightpwm;

	EX0=1;
	EX1=1;
}



void EXT0_ISR (void) interrupt 0
{
	unsigned short int tempt2ofcount=t2overflowcount; 
	unsigned int timer2val=TMR2;//grab timer 2 value

	
	IE0 = 0;
	//update interrupt period

	leftencoderperiod=(unsigned long int) ((timer2val/400) + (t2count/400)*tempt2ofcount);	//total time in .1 ms units
	if(leftencoderperiod<leftencoderoffset)	//timer 2 overflowed, adjust for it
	{
		leftencoderperiod+=~ ((unsigned long int)(0)) - leftencoderoffset;
	}
	else
	{//subtract off offset for the time between last interrupt
		leftencoderperiod-=leftencoderoffset;
	}

	leftencoderoffset=(unsigned long int)((timer2val/400) + (t2count/400)*tempt2ofcount);

	if(desired_left_pwm==0)
	{	//left motor shut off, make sure pwm is 0%
		PCA0CP0=0xffff;
	}
	else
	{
		contrlvar+=leftinc;
		encoder_pwmupdate();
	}
} 

void EXT1_ISR (void) interrupt 2
{
	unsigned short int tempt2ofcount=t2overflowcount; 
	unsigned int timer2val=TMR2;//grab timer 2 value
	
	
	IE1 = 0;
	//update interrupt period

	rightencoderperiod=(unsigned long int) ((timer2val/400) + (t2count/400)*tempt2ofcount);	//total time in .1 ms units
	if(rightencoderperiod<rightencoderoffset)	//timer 2 overflowed, adjust for it
	{
		rightencoderperiod+= ~((unsigned long int)(0)) - rightencoderoffset;
	}
	else
	{//subtract off offset for the time between last interrupt
		rightencoderperiod-=rightencoderoffset;
	}

	rightencoderoffset=(unsigned long int)((timer2val/400) + (t2count/400)*tempt2ofcount);

	
	if(desired_right_pwm==0)
	{
		PCA0CP1=0xffff;
	}
	else
	{
		contrlvar-=rightinc;
		encoder_pwmupdate();
	}
}


void SPI0_ISR (void) interrupt 6
{//compass reading program.
	//note: compass operating in binary mode, first 7 bits are useless, and the remaining 9 bits are data.
	unsigned short int readbyte=0;
	readbyte=SPI0DAT;	//read current byte

	SPIF =0;	//clear all interrupt flags
	WCOL =0;	//does not check for errors, since unsure how to correct them in this code
	MODF =0;
	RXOVRN =0;

	if(bytenumber==0)
	{//first recieved byte, save it and return
		firstbyte=readbyte;
		bytenumber=1;
	}
	else
	{
		secondbyte=readbyte;
		bytenumber=0;
		vehicledirection = (0x01&firstbyte)*0x0100 + secondbyte;  //combine together the two bytes of information
		vehicledirection&=0x01FF;		//mask off unused bits
		flags|=0x08; //sets compass reading flag
	}
}